c This code minimizes resonant field errors and normal field errors at plasma boundary
c The variables are typically either trim coil currents or realignment DOF's
c It uses NLP as an alternative to the svd solution which targeted only resonant fields.
c**************************************************************************************
c trim_nlp3.f - modified from trim_nlp.f to add targeting of b.n on plasma
c               uses bdotn file created by vacisld9+
c trim_nlp2.f - from trim_nlp.f : uses analytic calculation of obj_grad
c trim_nlp.f  - from trim_svd.f : replaces svd solve of Ax=b with nag nlp routine
c***********************************************************
      module svddata
      implicit real*8(a-h,o-z)
      real*8, dimension(:,:), allocatable :: bmns_coils
      real*8, dimension(:), allocatable :: bmns_target, bmns_resid, bmns
      real*8, dimension(:), allocatable :: curs,x_upper,x_lower
      real*8, dimension(:,:), allocatable :: bdotn_coils
      real*8, dimension(:), allocatable :: bdotn_target, bdotn_resid
      real*8 :: curmax, bmns1, bmns_resid2, bdotn_max, bdotn_resid2
      real*8 :: wbmns, wbdotn_max, wbdotn_resid2
       real*8 :: obj_f0, obj_f1
      integer :: mmode, mcoil
      character*32 lab
      character*16, dimension(:), allocatable ::  lab_target, lab_coil
      integer, dimension(:), allocatable :: m
      real*8, dimension(:), allocatable :: xiotap, wm
      end module svddata
c***********************************************************
      use svddata
      use nag_nlp
      implicit real*8(a-h,o-z)
      type(nag_nlp_cntrl_dp) :: cntrl
      character*16 labx, ext
      external obj_fun, con_fun
c
c get job name from command line
c
      iarg=iargc()
      if(iarg.lt.1)stop " usage: ./trim_nlp3.x ext"
      narg = getarg(1,ext)
c
      call nag_nlp_cntrl_init(cntrl)
      cntrl.list=.false.
      cntrl.unit=15
       cntrl.major_iter_lim=1000
      cntrl.optim_tol=1.d-4
c
c Read in Control Data and allocate arrays
c
      open(unit=9,file="trim_nlp3.inp"//trim(ext),status="old")
      read(9, *) mmode, mcoil, npts ! model size info
c
      allocate(bmns_coils(mmode,mcoil)) ! A matrix
      allocate(lab_target(mmode))
      allocate(lab_coil(mcoil))
      allocate(curs(mcoil), bmns_target(mmode), bmns_resid(mmode))
      allocate(bmns(mmode))
      allocate(bdotn_target(npts))
      allocate(bdotn_resid(npts))
      allocate(bdotn_coils(npts,mcoil))
      allocate(m(mmode))
      allocate(wm(mmode))
      allocate(xiotap(mmode))
      allocate(x_upper(mcoil),x_lower(mcoil))
c
      read(9,*) wbmns, wbdotn_max, wbdotn_resid2, curmax
      read(9,*) m(1:mmode)
      read(9,*) wm(1:mmode)
      read(9,*) xiotap(1:mmode)
      read(9,*) (lab_target(i),i=1,mmode)
      close(9)
c
c
c Read in  Bmn data from Vacisld9+
c
      open(unit=7,file="bmns"//trim(ext),status="old")
      read(7,*) labx, bmns_target
c
      do j=1,mcoil
      read(7,*) lab_coil(j), (bmns_coils(i,j),i=1,mmode)
      enddo
      close(7)
c
c
c Read in B.n data from Vacisld9+
c
      open(unit=11,file="bdotn"//trim(ext),status="old")
      read(11,*)
      read(11,*) bdotn_target(1:npts)
      bdotn_target=-bdotn_target
      do icoil=1,mcoil
        read(11,*)
        read(11,*) bdotn_coils(1:npts,icoil)
      enddo
      close(11)
c
c Evaluate initial object function
c
      curs = 0.d0
      bmns_resid=matmul(bmns_coils,curs)-bmns_target
      bmns_resid2=dot_product(bmns_resid*wm,bmns_resid*wm)*1.d12
      bdotn_resid=matmul(bdotn_coils,curs)-bdotn_target
      bdotn_resid2=dot_product(bdotn_resid,bdotn_resid)*1.d12
      bdotn_max = maxval(abs(bdotn_resid))
      obj_f0=bmns_resid2*wbmns + bdotn_max*wbdotn_max
     +         + bdotn_resid2*wbdotn_resid2
c
      do j=1,mmode,2
       bmns(j) = sqrt( bmns_target(j)**2 + bmns_target(j+1)**2)
        bmns(j) = 4.d0*sqrt(abs(bmns(j))/m(j)/xiotap(j))
      enddo
      call sum_island2(bmns, xiotap, m, bmns1)
c
      write(*,*) "***** Initial *****"
      write(*,*) "Max Current = ", maxval(abs(curs))
      write(*,*) "Total Islands = ", bmns1*100.d0
      write(*,*) "bdotn_max   = ", bdotn_max
      write(*,*) "bdotn_avg      = ", sqrt(bdotn_resid2*1.d-12/npts)
      write(*,*)
      write(*,*) "bmns_resid2 = ", bmns_resid2
      write(*,*) "bdotn_resid2   = ", bdotn_resid2
      write(*,*) "obj_f0   = ", obj_f0
c
      write(*,*)
      write(*,*) "curs"
      write(*,"(1p6e15.6)") curs
      write(*,"(6a15)") lab_target
      write(*,*) "bmns_target"
      write(*,"(1p6e15.6)") bmns_target
c
      write(*,*) "Island Size, %Total Flux"
      write(*,"(6a15)") lab_target(1:mmode:2)
      write(*,"(6f15.2)") bmns(1:mmode:2)*100.d0, bmns1*100.d0
c
c Do NLP solution (non-linear programming)
c
      bmns_target=-bmns_target
      x_lower(1:mcoil)=-curmax
      x_upper(1:mcoil)=+curmax
      call nag_nlp_sol(obj_fun,curs,obj_f,obj_deriv=.false.,    !  obj deriv=.true.
     +   con_deriv=.false., num_nlin_con=0, con_fun=con_fun,
     + control=cntrl,
     +   x_lower=x_lower, x_upper=x_upper)
c
      obj_f1=bmns_resid2*wbmns + bdotn_max*wbdotn_max
     +         + bdotn_resid2*wbdotn_resid2
c
      do j=1,mmode,2
       bmns(j) = sqrt( bmns_resid(j)**2 + bmns_resid(j+1)**2)
        bmns(j) = 4.d0*sqrt(abs(bmns(j))/m(j)/xiotap(j))
      enddo
      call sum_island2(bmns, xiotap, m, bmns1)
c
      write(*,*) "***** Solution *****"
      write(*,*) "Max Current = ", maxval(abs(curs))
      write(*,*) "Total Islands = ", bmns1*100.d0
      write(*,*) "bdotn_max   = ", bdotn_max
      write(*,*) "bdotn_avg      = ", sqrt(bdotn_resid2*1.d-12/npts)
      write(*,*)
      write(*,*) "bmns_resid2 = ", bmns_resid2
      write(*,*) "bdotn_resid2   = ", bdotn_resid2
      write(*,*) "obj_f1   = ", obj_f1
      write(*,*)
      write(*,*) "curs"
      write(*,"(1p6e15.6)") curs
      write(*,"(6a15)") lab_target
      write(*,*) "bmns_resid"
      write(*,"(1p6e15.6)") bmns_resid
c
      write(*,*) "Island Size, %Total Flux"
      write(*,"(6a15)") lab_target(1:mmode:2)
      write(*,"(6f15.2)") bmns(1:mmode:2)*100.d0, bmns1*100.d0
      stop
      end
c***********************************************************
      subroutine obj_fun(first_call, x, finish, obj_f, obj_grad)
      use svddata
      implicit real*8(a-h,o-z)
c
      logical, intent(in) :: first_call
      real*8, intent(in) :: x(:)
      logical, intent(out) :: finish
      real*8, intent(out) :: obj_f
      real*8, intent(inout), optional :: obj_grad(:)
c
      bmns_resid=matmul(bmns_coils,x)-bmns_target
       bmns_resid2=dot_product(bmns_resid*wm,bmns_resid*wm)*1.d12
!      obj_grad=2.d0*matmul(bmns_resid,bmns_coils)*1.d12
c
c add max(b.n) to object function
c
      bdotn_resid=matmul(bdotn_coils,x)-bdotn_target
       bdotn_resid2=dot_product(bdotn_resid,bdotn_resid)*1.d12
      bdotn_max = maxval(abs(bdotn_resid))
c
      obj_f=bmns_resid2*wbmns + bdotn_max*wbdotn_max
     +         + bdotn_resid2*wbdotn_resid2
c
      return
      end
      subroutine con_fun
      return
      end
c***********************************************************
      subroutine sum_island2(bmns, xiotap, m, bmns1)
      implicit none
      integer, parameter :: nmns = 20
      real*8, dimension(nmns) :: bmns
      real*8, dimension(nmns) :: xiotap
      integer, dimension(nmns) :: m
      real*8 ::  ds, bmns1
      integer :: mi, nm
c
c this routine sums islands assuming the following resonances are present
c IT DOES NOT WORK FOR ANY OTHER  LIST OF RESONANCES
c                  m5n3s   m5n3c  m10n6s  m10n6c   m6n3s   m6n3c
c                 m12n6s  m12n6c   m7n3s   m7n3c  m14n6s  m14n6c
c                  m2n1s   m2n1c   m4n2s   m4n2c   m3n2s   m3n2c
c                  m6n4s   m6n4c
c
c 4 surfaces in total, 2 resonances each except 1/2 whcih combines with 3/6
c
      real*8, dimension(4) :: bms, bmc
c
      bmns1 = 0.d0 ! initialize sum
c
c 3/5 surface
c
      ds=max(bmns(1),bmns(3))
      bmns1 = bmns1 + ds
c
c 1/2 surface & 3/6
c
      ds=max(bmns(13),bmns(15),bmns(5),bmns(7))
      bmns1 = bmns1 + ds
c
c 3/7 surface
c
      ds=max(bmns(9),bmns(11))
      bmns1 = bmns1 + ds
c
c 2/3 surface
c
      ds=max(bmns(17),bmns(19))
      bmns1 = bmns1 + ds
c
      return
      end
c***********************************************************
